/*
** pork_screen_cmd.c - screen management.
** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** The functions in this file are mostly wrappers around
** functions that do the actual work.  The difference is
** these functions do any necessary updating of the contents
** of the screen (i.e. refreshing, redrawing, etc).
*/

#include <config.h>

#include <unistd.h>
#include <ncurses.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>

#include <libfaim/aim.h>
#include <libfaim/faimconfig.h>

#include <pork.h>
#include <pork_missing.h>
#include <pork_util.h>
#include <pork_list.h>
#include <pork_buddy.h>
#include <pork_set.h>
#include <pork_swindow.h>
#include <pork_imwindow.h>
#include <pork_slist.h>
#include <pork_buddy_list.h>
#include <pork_acct.h>
#include <pork_cstr.h>
#include <pork_misc.h>
#include <pork_input.h>
#include <pork_bind.h>
#include <pork_events.h>
#include <pork_screen.h>
#include <pork_screen_cmd.h>
#include <pork_status.h>
#include <pork_aim.h>

extern struct screen screen;

inline void screen_draw_cursor(void) {
	struct imwindow *imwindow = cur_window();

	wmove(screen.status_bar, STATUS_ROWS - 1,
		input_get_cursor_pos(imwindow->input));
	wnoutrefresh(screen.status_bar);
}

inline void screen_doupdate(void) {
	int cur_old = curs_set(0);

	screen_draw_cursor();
	doupdate();
	curs_set(cur_old);
}

inline struct imwindow *cur_window(void) {
	return (screen.cur_window->data);
}

int screen_draw_input(void) {
	struct imwindow *imwindow = cur_window();

	if (imwindow->input->dirty) {
		char *input_line;

		wmove(screen.status_bar, STATUS_ROWS - 1, 0);
		wclrtoeol(screen.status_bar);

		input_line = input_partial(imwindow->input);
		if (input_line == imwindow->input->input_buf &&
			imwindow->input->prompt != NULL)
		{
			wputstr(screen.status_bar, imwindow->input->prompt);
		}

		waddnstr(screen.status_bar,
			input_partial(imwindow->input), imwindow->input->width);

		/*
		** We don't need a wnoutfresh here for this window.  Since it's
		** dirty (implying we return 1), screen_doupdate() will be called,
		** which redraws the cursor and calls wnoutrefresh on the status
		** window, which happens to be the window the input is drawn in.
		*/

		imwindow->input->dirty = 0;
		return (1);
	}

	return (0);
}

void screen_window_swap(dlist_t *new_cur) {
	struct imwindow *imwindow = NULL;
	u_int32_t last_own_input = 0;
	u_int32_t cur_own_input;
	struct aim_priv *priv;

	if (screen.cur_window != NULL) {
		imwindow = cur_window();

		last_own_input = wopt_get_bool(imwindow->opts, WOPT_PRIVATE_INPUT);
		imwindow->swindow.visible = 0;
	}

	screen.cur_window = new_cur;

	imwindow = cur_window();
	imwindow->swindow.visible = 1;
	imwindow->swindow.activity = 0;
	cur_own_input = wopt_get_bool(imwindow->opts, WOPT_PRIVATE_INPUT);

	/*
	** If the current window and the one that we're going to switch
	** to don't share an input buffer, history, etc, the input
	** line must be refreshed.  Force a redraw.
	*/

	if (cur_own_input || last_own_input) {
		imwindow->input->dirty = 1;
		screen_draw_input();
	}

	status_draw(imwindow->priv);

	priv = imwindow->priv;

	/*
	** To force ncurses to redraw it on the physical screen.
	*/
	redrawwin(imwindow->swindow.win);
	imwindow->swindow.dirty = 1;
	imwindow_blist_draw(imwindow);
}

inline int screen_goto_window(u_int32_t refnum) {
	dlist_t *cur = screen_find_refnum(refnum);

	if (cur == NULL)
		return (-1);

	screen_window_swap(cur);
	return (0);
}

void screen_refresh(void) {
	struct imwindow *imwindow = cur_window();

	wclear(curscr);

	wclear(imwindow->swindow.win);
	swindow_redraw(&imwindow->swindow);
	imwindow_blist_draw(imwindow);

	status_draw(imwindow->priv);

	imwindow_refresh(imwindow);

	imwindow->input->dirty = 1;
	screen_draw_input();

	screen_doupdate();
}

struct imwindow *screen_new_window(	struct aim_priv *priv,
									char *target,
									char *name)
{
	u_int32_t refnum = screen_get_new_refnum();
	struct imwindow *imwindow;

	imwindow = imwindow_new(screen.rows - STATUS_ROWS, screen.cols,
		refnum, TYPE_IM, priv, target);
	if (imwindow == NULL)
		return (NULL);

	if (name != NULL)
		imwindow_rename(imwindow, name);

	screen_add_window(imwindow);
	status_draw(imwindow->priv);

	return (imwindow);
}

struct imwindow *screen_new_chat_window(struct aim_priv *priv, char *name) {
	u_int32_t refnum = screen_get_new_refnum();
	struct imwindow *imwindow;

	imwindow = imwindow_new(screen.rows - STATUS_ROWS, screen.cols,
		refnum, TYPE_CHAT, priv, name);
	if (imwindow == NULL)
		return (NULL);

	imwindow->chat = NULL;

	screen_add_window(imwindow);
	status_draw(imwindow->priv);

	return (imwindow);
}

inline void screen_cycle_fwd(void) {
	screen_window_swap(screen.cur_window->next);
}

inline void screen_cycle_bak(void) {
	screen_window_swap(screen.cur_window->prev);
}

int screen_close_window(struct imwindow *imwindow) {
	dlist_t *node = screen_find_refnum(imwindow->refnum);

	if (node == NULL)
		return (-1);

	/*
	** The status window can't be closed.
	*/

	if (imwindow->type == TYPE_STATUS)
		return (-1);

	if (imwindow->type == TYPE_CHAT && imwindow->chat != NULL) {
		aim_leave_chatroom(imwindow->priv, imwindow->chat->title,
			imwindow->chat->exchange);
	}

	if (imwindow->swindow.visible)
		screen_cycle_fwd();

	screen_window_list_remove(node);
	imwindow_destroy(imwindow);
	free(node);

	return (0);
}

inline int screen_set_quiet(int status) {
	int ret = screen.quiet;

	screen.quiet = status;
	return (ret);
}

static void __screen_win_msg(	struct imwindow *win,
								int timestamp,
								int show_banner,
								int color,
								char *fmt,
								va_list ap)
{
	char buf[4096];
	chtype ch[4096];
	size_t chlen = array_elem(ch);
	size_t banner_len;
	char *banner;

#ifndef ENABLE_DEBUGGING
	if (win == NULL)
		return;
#endif

	if (show_banner) {
		banner = opt_get_str(OPT_BANNER);

		banner_len = plaintext_to_cstr(banner, ch, chlen);
		chlen -= banner_len;
	} else
		banner_len = 0;

	vsnprintf(buf, sizeof(buf), fmt, ap);

	if (color)
		plaintext_to_cstr(buf, &ch[banner_len], chlen);
	else
		plaintext_to_cstr_nocolor(buf, &ch[banner_len], chlen);

	swindow_add(&win->swindow, ch, timestamp);
}

inline void screen_win_msg(	struct imwindow *win,
							int ts,
							int banner,
							int color,
							char *fmt,
							...)
{
	va_list ap;

	va_start(ap, fmt);
	__screen_win_msg(win, ts, banner, color, fmt, ap);
	va_end(ap);
}

inline void screen_status_msg(char *fmt, ...) {
	va_list ap;

	va_start(ap, fmt);
	if (!screen.quiet)
		__screen_win_msg(cur_window(), 0, 1, 1, fmt, ap);
	va_end(ap);
}

inline void screen_err_msg(char *fmt, ...) {
	va_list ap;

	va_start(ap, fmt);
	__screen_win_msg(cur_window(), 0, 1, 1, fmt, ap);
	va_end(ap);
}

inline void screen_output(char *fmt, ...) {
	va_list ap;

	va_start(ap, fmt);
	__screen_win_msg(cur_window(), 0, 1, 1, fmt, ap);
	va_end(ap);
}

inline void screen_nocolor_msg(char *fmt, ...) {
	va_list ap;

	va_start(ap, fmt);
	if (!screen.quiet)
		__screen_win_msg(cur_window(), 0, 1, 0, fmt, ap);
	va_end(ap);
}

void screen_win_target_msg(char *target, int timestamp, char *fmt, ...) {
	va_list ap;
	struct imwindow *win = imwindow_find(target);

	if (win == NULL)
		win = cur_window();

	va_start(ap, fmt);
	__screen_win_msg(win, timestamp, 0, 1, fmt, ap);
	va_end(ap);
}

void screen_print_info(	char *user,
						char *profile,
						char *away_msg,
						u_int32_t warn_level,
						u_int32_t idle_time,
						u_int32_t online_since,
						u_int32_t member_since,
						int print_all)
{
	struct imwindow *imwindow = cur_window();
	char buf[2048];
	chtype ch[2048];
	size_t chlen = array_elem(ch);

	if (print_all) {
		snprintf(buf, sizeof(buf),
			"-%%p-%%P---%%Cu%%csername%%W:%%w %s", user);

		plaintext_to_cstr(buf, ch, chlen);
		swindow_add(&imwindow->swindow, ch, 0);

		if (idle_time > 0) {
			char time_buf[64];

			time_to_str(idle_time, time_buf, sizeof(time_buf));

			snprintf(buf, sizeof(buf),
				"-%%p-%%P--%%Ci%%cdle time%%W:%%w %s", time_buf);

			plaintext_to_cstr(buf, ch, chlen);
			swindow_add(&imwindow->swindow, ch, 0);
		}

		if (warn_level > 0) {
			snprintf(buf, sizeof(buf), "-%%p-%%P-%%Cw%%carn level%%W:%%w %u%%",
				warn_level);
			plaintext_to_cstr(buf, ch, chlen);
			swindow_add(&imwindow->swindow, ch, 0);
		}

		if (online_since) {
			snprintf(buf, sizeof(buf), "-%%Co%%cnline since%%W:%%w %s",
				asctime(localtime((time_t *) &online_since)));
			plaintext_to_cstr(buf, ch, chlen);
			swindow_add(&imwindow->swindow, ch, 0);
		}

		if (member_since) {
			snprintf(buf, sizeof(buf), "-%%Cm%%cember since%%W:%%w %s",
				asctime(localtime((time_t *) &online_since)));
			plaintext_to_cstr(buf, ch, chlen);
			swindow_add(&imwindow->swindow, ch, 0);
		}

		if (profile) {
			if (strchr(profile, '\n') == NULL) {
				snprintf(buf, sizeof(buf),
					"-%%p-%%P--%%Cu%%cser info%%W:%%w %s", profile);

				plaintext_to_cstr(buf, ch, chlen);
				swindow_add(&imwindow->swindow, ch, 0);
			} else {
				snprintf(buf, sizeof(buf), "-%%p-%%P--%%Cu%%cser info%%W:");
				plaintext_to_cstr(buf, ch, chlen);
				swindow_add(&imwindow->swindow, ch, 0);
				screen_print_multiline(imwindow, profile, 0);
			}
		}
	}

	if (away_msg) {
		if (strchr(away_msg, '\n') == NULL) {
			snprintf(buf, sizeof(buf), "-%%Ca%%cway message%%W:%%w %s",
				away_msg);
			plaintext_to_cstr(buf, ch, chlen);
			swindow_add(&imwindow->swindow, ch, 0);
		} else {
			snprintf(buf, sizeof(buf), "-%%Ca%%cway message%%W:");
			plaintext_to_cstr(buf, ch, chlen);
			swindow_add(&imwindow->swindow, ch, 0);
			screen_print_multiline(imwindow, away_msg, 0);
		}
	}
}

/*
** Print out a string that may have newlines
** embedded in it to the specified window.
*/

void screen_print_multiline(struct imwindow *win,
							char *msg,
							int timestamp)
{
	size_t len = strlen(msg) + 8;
	chtype *buf = xmalloc(sizeof(chtype) * len);
	char *p;

	p = msg;
	while (p != NULL) {
		char *next = strchr(p, '\n');
		if (next != NULL)
			*next++ = '\0';

		buf[0] = buf[1] = ' ';
		plaintext_to_cstr(p, &buf[2], len - 2);
		swindow_add(&win->swindow, buf, timestamp);
		p = next;
	}

	free(buf);
}

int screen_get_passwd(char *buf, size_t len) {
	int ret;

	buf[0] = '\0';

	wmove(screen.status_bar, 1, 0);
	wclrtoeol(screen.status_bar);

	waddstr(screen.status_bar, "Password: ");
	wrefresh(screen.status_bar);

	ret = wgetnstr(screen.status_bar, buf, len);
	return (ret);
}
